using System;
using System.Linq;
using System.Collections.Generic;

// A LayeredInterpolator consists of one base interpolator as well as an arbitrary
// amount of interpolators which are stacked on top. Only the interpolator at the
// top of the queue will be used. Once that interpolator has finished, it will be
// removed from the queue and the seond-last interpolator takes over.
[Serializable]
public class InterpolatorQueue
{

    // Should probably be a stack but whatever
    private List<Interpolator> interpolators = new List<Interpolator>();

    public bool hasInterpolatorsQueued
    {
        get => interpolators.Count >= 2;
    }

    public bool running
    {
        get => interpolators.Any(i => i.running);
    }
    public float currentValue
    {
        get => GetCurrentInterpolator().currentValue;
        set => GetCurrentInterpolator().currentValue = value;
    }
    public float targetValue
    {
        get => GetCurrentInterpolator().targetValue;
        set => GetCurrentInterpolator().targetValue = value;
    }
    public float velocity
    {
        get => GetCurrentInterpolator().velocity;
        set => GetCurrentInterpolator().velocity = value;
    }

    public InterpolatorQueue(Interpolator baseInterpolator)
    {
        interpolators.Add(baseInterpolator);
    }

    public void PushInterpolator(Interpolator newInterpolator)
    {
        // Only push interpolators that would actually do something
        if (newInterpolator.running)
        {
            interpolators.Add(newInterpolator);
        }
    }

    public void PopInterpolator()
    {
        // Always keep the base interpolator
        if (interpolators.Count >= 2)
        {
            interpolators.RemoveAt(interpolators.Count - 1);
        }
    }

    public float Tick()
    {
        Interpolator current = GetCurrentInterpolator();
        float tickValue = current.Tick();
        for (var i = 0; i < interpolators.Count - 1; i++)
        {
            var interpolator = interpolators[i];
            // Set velocity and currentValoe of all other interpolators to allow smooth transitions
            interpolator.velocity = current.velocity;
            interpolator.currentValue = tickValue;
            if (i == 0)
            {
                // Set target of base interpolator to avoid snap-back when all other interpolators have finished
                interpolator.targetValue = tickValue;
            }
        }
        // Remove non-running interpolators from the top of the queue
        while (!GetCurrentInterpolator().running && interpolators.Count >= 2)
        {
            PopInterpolator();
        }
        return tickValue;
    }

    private Interpolator GetCurrentInterpolator()
    {
        return interpolators.Last();
    }

}
